import torch
import torch.nn as nn

from torch import Tensor
from typing import *

from core.distributed import PGraph
from torch_geometric.nn import Linear
import torch_geometric.nn.inits as pyg_inits

class GCNConv(nn.Module):
    def __init__(self,
        in_channels: int,
        out_channels: int,
    ) -> None:
        super().__init__()
        self.linear = Linear(in_channels, out_channels, bias=False, weight_initializer="glorot")
        self.bias = nn.Parameter(torch.empty(out_channels))
        self.reset_parameters()
        
    def reset_parameters(self):
        self.linear.reset_parameters()
        pyg_inits.zeros(self.bias)
    
    def forward(self,
        g: PGraph,
        x: Tensor,
        edge_attr: Tensor,
    ) -> Tensor:
        x = self.linear(x)
        x = g.gather(x, src=True, dst=False)
        x = x * edge_attr.view(-1, 1)
        x = g.scatter(x, src=False, dst=True, reduce_op="sum")
        return x + self.bias
    
